# flake8: noqa: F401
from .base import AuthManager
from .basic_auth import (
    BasicAuthManager,
    BasicAuthHandler,
    manager as basic_auth_manager,
    HTTP_BASIC_AUTH_NAME,
)
from .token import (
    TokenAuthManager,
    TokenAuthHandler,
    manager as token_auth_manager,
    TOKEN_NAME,
)



from conda.cli.conda_argparse import BUILTIN_COMMANDS

from conda_auth import plugin
from conda_auth.constants import PLUGIN_NAME
from conda_auth.handlers import (
    HTTP_BASIC_AUTH_NAME,
    TOKEN_NAME,
    BasicAuthHandler,
    TokenAuthHandler,
)


def test_conda_subcommands_hook():
    """
    Test to make sure that this hook yields the correct objects.
    """
    objs = list(plugin.conda_subcommands())

    assert objs[0].name == "auth"
    assert objs[0].summary == "Authentication commands for conda"


def test_conda_pre_commands_hook():
    """
    Test to make sure that this hook yields the correct objects.
    """
    objs = list(plugin.conda_pre_commands())

    run_for = BUILTIN_COMMANDS.union(plugin.ENV_COMMANDS)

    assert objs[0].name == f"{PLUGIN_NAME}-{HTTP_BASIC_AUTH_NAME}"
    assert objs[0].run_for == run_for

    assert objs[1].name == f"{PLUGIN_NAME}-{TOKEN_NAME}"
    assert objs[1].run_for == run_for


def test_conda_auth_handlers_hook():
    """
    Test to make sure that this hook yields the correct objects.
    """
    objs = list(plugin.conda_auth_handlers())

    assert objs[0].name == HTTP_BASIC_AUTH_NAME
    assert objs[0].handler == BasicAuthHandler

    assert objs[1].name == TOKEN_NAME
    assert objs[1].handler == TokenAuthHandler



from conda.exceptions import CondaError


class CondaAuthError(CondaError):
    """Custom error for the conda-auth plugin"""


class InvalidCredentialsError(CondaAuthError):
    """Error raised when credentials are invalid"""



"""
A place to register plugin hooks
"""
from conda.cli.conda_argparse import BUILTIN_COMMANDS
from conda.plugins import CondaAuthHandler, CondaPreCommand, CondaSubcommand, hookimpl

from .handlers import (
    basic_auth_manager,
    token_auth_manager,
    BasicAuthHandler,
    TokenAuthHandler,
    HTTP_BASIC_AUTH_NAME,
    TOKEN_NAME,
)
from .cli import auth
from .constants import PLUGIN_NAME

ENV_COMMANDS = {
    "env_config",
    "env_create",
    "env_export",
    "env_list",
    "env_remove",
    "env_update",
}


@hookimpl
def conda_subcommands():
    """
    Registers subcommands
    """
    yield CondaSubcommand(
        name="auth",
        action=lambda args: auth(prog_name="conda auth", args=args),
        summary="Authentication commands for conda",
    )


@hookimpl
def conda_pre_commands():
    """
    Registers pre-command hooks
    """
    yield CondaPreCommand(
        name=f"{PLUGIN_NAME}-{HTTP_BASIC_AUTH_NAME}",
        action=basic_auth_manager.hook_action,
        run_for=BUILTIN_COMMANDS.union(ENV_COMMANDS),
    )
    yield CondaPreCommand(
        name=f"{PLUGIN_NAME}-{TOKEN_NAME}",
        action=token_auth_manager.hook_action,
        run_for=BUILTIN_COMMANDS.union(ENV_COMMANDS),
    )


@hookimpl
def conda_auth_handlers():
    """
    Registers auth handlers
    """
    yield CondaAuthHandler(name=HTTP_BASIC_AUTH_NAME, handler=BasicAuthHandler)
    yield CondaAuthHandler(name=TOKEN_NAME, handler=TokenAuthHandler)



"""
Module for custom click.Option classes
"""
from __future__ import annotations
import click
from typing import Any
from collections.abc import Mapping


class ConditionalOption(click.Option):
    """
    Custom option that does the following things:

    - Define ``mutually_exclusive`` options that cannot be passed together
    - Control prompting in the presence of other options via ``prompt_when``
    """
    def __init__(self, *args, **kwargs):
        self.not_required_if = set(kwargs.pop("not_required_if", []))

        self.prompt_when = set(kwargs.pop("prompt_when", []))
        if self.prompt_when:
            # ensure prompt text is configured,
            # conditionally control whether we prompt in handle_parse_result
            kwargs["prompt"] = True

        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        if self.mutually_exclusive:
            # augment help blurb to include details about mutually exclusive options
            help_ = kwargs.get("help", "")
            mutex = ", ".join(map(repr, self.mutually_exclusive))
            kwargs["help"] = f"{help_}; cannot be used with these options: {mutex}"

        super().__init__(*args, **kwargs)

    def handle_parse_result(
        self,
        ctx: click.Context,
        opts: Mapping[str, Any],
        args: list[str],
    ) -> tuple[Any, list[str]]:
        # determine whether mutex has been violated
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            mutex = ", ".join(map(repr, self.mutually_exclusive))
            raise click.UsageError(f"Option {self.name!r} cannot be used with {mutex}")

        # determine whether we want to prompt for this argument
        if self.prompt_when and not self.prompt_when.intersection(opts):
            self.prompt = None

        if (
            self.not_required_if
            and self.name not in opts
            and not self.not_required_if.intersection(opts)
        ):
            required = {self.name, *self.not_required_if}
            raise click.MissingParameter(
                ctx=ctx,
                param_type="option",
                param_hint=" / ".join(sorted(map(repr, required))),
            )

        return super().handle_parse_result(ctx, opts, args)



import pytest
from click.testing import CliRunner


@pytest.fixture
def keyring(mocker):
    """
    Used to mock keyring for the duration of our tests
    """

    def _keyring(secret):
        get_keyring = mocker.patch("conda_auth.storage.get_keyring")
        keyring_storage = mocker.patch("conda_auth.storage.keyring.keyring")
        keyring_storage.get_password.return_value = secret

        return keyring_storage, get_keyring

    return _keyring


@pytest.fixture
def runner():
    """
    CLI test runner used for all tests
    """
    yield CliRunner()


@pytest.fixture
def condarc(mocker):
    """
    Mocks the CondaRC object
    """
    condarc_mock = mocker.patch("conda_auth.cli.CondaRC")

    return condarc_mock



from conda_auth.cli import auth, SUCCESSFUL_LOGIN_MESSAGE
from conda_auth.condarc import CondaRCError
from conda_auth.exceptions import CondaAuthError


def test_login_basic_auth_no_options(runner, keyring, condarc):
    """
    Runs the login command with no additional CLI options defined (e.g. --username)
    """
    username = "user"
    secret = "password"
    channel_name = "tester"

    # setup mocks
    keyring(None)

    # run command
    result = runner.invoke(
        auth, ["login", channel_name, "--basic"], input=f"{username}\n{secret}"
    )

    assert result.exit_code == 0
    assert SUCCESSFUL_LOGIN_MESSAGE in result.output


def test_login_with_options_basic_auth(runner, keyring, condarc):
    """
    Runs the login command with CLI options defined (e.g. --username)
    """
    channel_name = "tester"

    # setup mocks
    keyring(None)

    # run command
    result = runner.invoke(
        auth,
        ["login", channel_name, "--basic", "--username", "test", "--password", "test"],
    )

    assert result.exit_code == 0
    assert SUCCESSFUL_LOGIN_MESSAGE in result.output


def test_login_with_invalid_auth_type(runner, keyring, condarc):
    """
    Runs the login command when there is an invalid auth type configured in settings
    """
    channel_name = "tester"

    # setup mocks
    keyring(None)

    # run command
    result = runner.invoke(auth, ["login", channel_name])
    exc_type, exception, _ = result.exc_info

    assert result.exit_code == 2
    assert exc_type == SystemExit
    assert "Error: Missing option 'basic' / 'token'." in result.stdout


def test_login_succeeds_error_returned_when_updating_condarc(runner, keyring, condarc):
    """
    Test the case where the login runs successfully but an error is returned when trying to update
    the condarc file.
    """
    channel_name = "tester"

    # setup mocks
    keyring(None)
    condarc().save.side_effect = CondaRCError("Could not save file")

    # run command
    result = runner.invoke(
        auth, ["login", channel_name, "--basic"], input="user\npassword"
    )
    exc_type, exception, _ = result.exc_info

    assert exc_type == CondaAuthError
    assert "Could not save file" == exception.message


def test_login_token(mocker, runner, keyring, condarc):
    """
    Test successful login with token
    """
    channel_name = "tester"

    # setup mocks
    mock_context = mocker.patch("conda_auth.cli.context")
    mock_context.channel_settings = []
    keyring(None)

    result = runner.invoke(auth, ["login", channel_name, "--token", "token"])

    assert result.exit_code == 0


def test_login_token_no_options(runner, keyring, condarc):
    """
    Test successful login with token without the value being supplied at the command line
    """
    channel_name = "tester"

    # setup mocks
    keyring(None)

    result = runner.invoke(auth, ["login", channel_name, "--token"], input="token\n")

    assert result.exit_code == 0
    assert SUCCESSFUL_LOGIN_MESSAGE in result.output



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping

import conda.base.context
from conda.models.channel import Channel
from conda.base.context import context as global_context

from ..storage import storage


class AuthManager(ABC):
    """
    Defines an interface for auth handlers to use within plugin
    """

    def __init__(
        self,
        context: conda.base.context.Context | None = None,
        cache: dict | None = None,
    ):
        """
        Optionally set a cache and context object to use
        """
        self._context = context or global_context
        self._cache = {} if cache is None else cache

    def hook_action(self, command: str) -> None:
        """Action callable to be used in the conda pre-command plugin hook"""

        for settings in self._context.channel_settings:
            if channel := settings.get("channel"):
                channel = Channel(channel)
                # Only attempt to authenticate for actively used channels
                if (
                    channel.canonical_name in self._context.channels
                    and settings.get("auth") == self.get_auth_type()
                ):
                    self.store(channel, settings)

    def store(self, channel: Channel, settings: Mapping[str, str]) -> str:
        """
        Used to retrieve credentials and store them on the ``cache`` property

        This method returns a "username" because this property could have been retrieved
        via user input while calling ``fetch_secret``.
        """
        extra_params = {
            param: settings.get(param) for param in self.get_config_parameters()
        }
        username, secret = self.fetch_secret(channel, extra_params)

        self.save_credentials(channel, username, secret)

        return username

    def save_credentials(self, channel: Channel, username: str, secret: str) -> None:
        """
        Saves the provided credentials to our credential store.
        """
        storage.set_password(self.get_keyring_id(channel), username, secret)

    def fetch_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> tuple[str, str]:
        """
        Fetch secrets and handle updating cache.
        """
        if secrets := self._cache.get(channel.canonical_name):
            return secrets

        secrets = self._fetch_secret(channel, settings)
        self._cache[channel.canonical_name] = secrets

        return secrets

    def get_secret(self, channel_name: str) -> tuple[str | None, str | None]:
        """
        Get the secret that is currently cached for the channel
        """
        secrets = self._cache.get(channel_name)

        if secrets is None:
            return None, None

        return secrets

    def cache_clear(self, channel_name: str | None = None) -> None:
        """
        Remove the internal cache for the manager object
        """
        if channel_name:
            self._cache.pop(channel_name, None)
        else:
            self._cache.clear()

    @abstractmethod
    def _fetch_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> tuple[str, str]:
        """Implementations should include routine for fetching secret"""

    @abstractmethod
    def remove_secret(self, channel: Channel, settings: Mapping[str, str]) -> None:
        """Implementations should include routine for removing secret"""

    @abstractmethod
    def get_auth_type(self) -> str:
        """
        Implementation should return a ``str`` which matches the name of the auth handler type in
        conda's configuration.
        """

    @abstractmethod
    def get_config_parameters(self) -> tuple[str, ...]:
        """
        Implementations should return a ``tuple`` of ``str`` which represent the configuration
        values to use in the ``context.channel_settings`` object.
        """

    @abstractmethod
    def get_keyring_id(self, channel: Channel) -> str:
        """
        Implementation should return the keyring id that will be used by the manager classes
        """

    @abstractmethod
    def get_auth_class(self) -> type:
        """
        Returns the authentication class to use (requests.auth.AuthBase subclass) for the given
        authentication manager
        """



"""
Basic auth implementation for the conda auth handler plugin hook
"""
from __future__ import annotations

from collections.abc import Mapping

from requests.auth import _basic_auth_str  # type: ignore
from conda.models.channel import Channel
from conda.plugins.types import ChannelAuthBase

from ..constants import PLUGIN_NAME
from ..exceptions import CondaAuthError
from ..storage import storage
from .base import AuthManager

USERNAME_PARAM_NAME: str = "username"
"""
Name of the configuration parameter where username information is stored
"""

PASSWORD_PARAM_NAME: str = "password"
"""
Name of the configuration parameter where password information is stored
"""

HTTP_BASIC_AUTH_NAME: str = "http-basic"
"""
Name used to refer to this authentication handler in configuration
"""


class BasicAuthManager(AuthManager):
    def get_keyring_id(self, channel: Channel):
        return f"{PLUGIN_NAME}::{HTTP_BASIC_AUTH_NAME}::{channel.canonical_name}"

    def _fetch_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> tuple[str, str]:
        """
        Gets the secrets by checking the keyring and then falling back to interrupting
        the program and asking the user for the credentials.
        """
        username = self.get_username(settings)
        password = self.get_password(username, settings, channel)

        return username, password

    def remove_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> None:
        keyring_id = self.get_keyring_id(channel)
        username = self.get_username(settings)

        storage.delete_password(keyring_id, username)

    def get_auth_type(self) -> str:
        return HTTP_BASIC_AUTH_NAME

    def get_config_parameters(self) -> tuple[str, ...]:
        return USERNAME_PARAM_NAME, PASSWORD_PARAM_NAME

    def get_username(self, settings: Mapping[str, str | None]):
        """
        Attempts to find username in settings and falls back to prompting user for it if not found.
        """
        username = settings.get(USERNAME_PARAM_NAME)

        if username is None:
            raise CondaAuthError("Username not found")

        return username

    def get_password(
        self, username: str, settings: Mapping[str, str | None], channel: Channel
    ) -> str:
        """
        Attempts to get password and falls back to prompting the user for it if not found.
        """
        # First see if a value has been passed in
        password = settings.get(PASSWORD_PARAM_NAME)

        # Now try retrieving it from the password manager
        if password is None:
            keyring_id = self.get_keyring_id(channel)
            password = storage.get_password(keyring_id, username)

            if password is None:
                raise CondaAuthError("Password not found")

        return password

    def get_auth_class(self) -> type:
        return BasicAuthHandler


manager = BasicAuthManager()


class BasicAuthHandler(ChannelAuthBase):
    """
    Implementation of HTTPBasicAuth that relies on a cache location for
    retrieving login credentials on object instantiation.

    Some of this has been copied over from ``requests.auth``.
    """

    def __init__(self, channel_name: str):
        super().__init__(channel_name)
        self.username, self.password = manager.get_secret(channel_name)

        if self.username is None and self.password is None:
            raise CondaAuthError(
                f"Unable to find user credentials for requests with channel {channel_name}"
            )

    def __eq__(self, other):
        return all(
            (
                self.username == getattr(other, "username", None),
                self.password == getattr(other, "password", None),
            )
        )

    def __ne__(self, other):
        return not self == other

    def __call__(self, request):
        request.headers["Authorization"] = _basic_auth_str(self.username, self.password)
        return request



from __future__ import annotations

import os
from pathlib import Path

from ruamel.yaml import YAML, YAMLError

yaml = YAML()


class CondaRCError(Exception):
    pass


class CondaRC:
    def __init__(self, condarc_path: Path | None = None):
        """
        Initializes the CondaRC object by attempting to open and load the contents
        of the condarc file found in the user's home directory.
        """
        self.condarc_path = condarc_path or Path(os.path.expanduser("~/.condarc"))

        try:
            self.condarc_path.touch()
            with self.condarc_path.open("r") as fp:
                contents = fp.read()
        except OSError as exc:
            raise CondaRCError(f"Could not open condarc file: {exc}")

        try:
            self.loaded_yaml = yaml.load(contents) or {}
        except YAMLError as exc:
            raise CondaRCError(f"Could not parse condarc: {exc}")

    def update_channel_settings(
        self, channel: str, auth_type: str, username: str | None = None
    ):
        """
        Update the condarc file's "channel_settings" section
        """
        if username is None:
            updated_settings = {"channel": channel, "auth": auth_type}
        else:
            updated_settings = {
                "channel": channel,
                "auth": auth_type,
                "username": username,
            }

        channel_settings = self.loaded_yaml.get("channel_settings", []) or []

        # Filter out the existing channel's entry if it's there
        filter_settings = [
            settings
            for settings in channel_settings
            if settings.get("channel") != channel
        ]

        # Add the updated settings map
        filter_settings.append(updated_settings)

        self.loaded_yaml["channel_settings"] = filter_settings

    def save(self):
        """Save the condarc file"""
        try:
            with self.condarc_path.open("w") as fp:
                yaml.dump(self.loaded_yaml, fp)
        except OSError as exc:
            raise CondaRCError(f"Could not save file: {exc}")



import pytest
from keyring.errors import NoKeyringError

from conda_auth.exceptions import CondaAuthError
from conda_auth.storage import get_storage_backend


def test_no_available_storage_backend(keyring):
    """
    Test to make sure we've covered the lines where an exception is raise
    if no storage backend can be found.
    """
    _, get_keyring_mock = keyring(None)

    get_keyring_mock.side_effect = NoKeyringError()

    with pytest.raises(
        CondaAuthError, match="Unable to find a credential storage backend"
    ):
        get_storage_backend()



from __future__ import annotations

import keyring
from keyring.errors import PasswordDeleteError

from .base import Storage
from ..exceptions import CondaAuthError


class KeyringStorage(Storage):
    """
    Storage implementation for keyring library
    """

    def get_password(self, key_id: str, username: str) -> str | None:
        return keyring.get_password(key_id, username)

    def set_password(self, key_id: str, username: str, password: str) -> None:
        return keyring.set_password(key_id, username, password)

    def delete_password(self, key_id: str, username: str) -> None:
        try:
            return keyring.delete_password(key_id, username)
        except PasswordDeleteError as exc:
            raise CondaAuthError(f"Unable to remove secret: {exc}")



# flake8: noqa: F401
from keyring import get_keyring
from keyring.errors import NoKeyringError

from .base import Storage
from .keyring import KeyringStorage
from ..exceptions import CondaAuthError


def get_storage_backend() -> Storage:
    """
    Determine the correct storage backend to use, raise CondaAuthError if none found.

    TODO: Add future support for another storage backend when keyring cannot be used.
    """
    try:
        keyring_tester = get_keyring()
        # Retrieve a dummy password to try to trigger NoKeyringError
        keyring_tester.get_password("conda_auth", "test")
    except NoKeyringError:
        raise CondaAuthError(
            "Unable to find a credential storage backend, which means this operating system"
            " is likely unsupported. One way to overcome this is by installing a third party"
            " keyring storage backend. You can find more information about that here: "
            "https://pypi.org/project/keyring"
        )

    return KeyringStorage()


storage = get_storage_backend()



from unittest.mock import MagicMock

import pytest
from ruamel.yaml.error import YAMLError

from conda_auth.condarc import CondaRC, CondaRCError, yaml


CONDARC_CONTENT = """
channels:
- defaults
channel_settings:
"""


def test_update_non_existing_condarc_file(tmp_path):
    """
    Make sure that the condarc file can be updated even when it first doesn't exist
    """
    channel = "tester"
    username = "username"
    auth_type = "http-basic"
    condarc_path = tmp_path / ".condarc"

    condarc = CondaRC(condarc_path)
    condarc.update_channel_settings(channel, auth_type, username)
    condarc.save()

    condarc_dict = yaml.load(condarc_path.read_text())

    assert condarc_dict == {
        "channel_settings": [
            {
                "channel": channel,
                "username": username,
                "auth": auth_type,
            }
        ]
    }


def test_update_existing_condarc_file(tmp_path):
    """
    Make sure that the condarc file can be updated even when it does exist

    TODO:
        It might be nice to expand this test in the future with some more condarc
        file states via a pytest.parameters decorator.
    """
    channel = "tester"
    username = "username"
    auth_type = "http-basic"
    condarc_path = tmp_path / ".condarc"
    condarc_path.write_text(CONDARC_CONTENT)

    condarc = CondaRC(condarc_path)
    condarc.update_channel_settings(channel, auth_type, username)
    condarc.save()

    condarc_dict = yaml.load(condarc_path.read_text())

    assert condarc_dict == {
        "channel_settings": [
            {
                "channel": channel,
                "username": username,
                "auth": auth_type,
            }
        ],
        "channels": ["defaults"],
    }


def test_error_while_reading_condarc_file():
    """
    Testing to make sure the appropriate error is raised when an OSError occurs
    """
    error_message = "Not allowed to read"
    mock_path = MagicMock()
    mock_path.open.side_effect = PermissionError(error_message)

    with pytest.raises(CondaRCError, match=error_message):
        CondaRC(mock_path)


def test_error_parsing_yaml(mocker, tmp_path):
    """
    Testing to make sure the appropriate error is raised when an YAMLError occurs
    """
    error_message = "Parse error"
    condarc_path = tmp_path / ".condarc"
    mock_yaml = mocker.patch("conda_auth.condarc.yaml")
    mock_yaml.load.side_effect = YAMLError(error_message)

    with pytest.raises(CondaRCError, match=error_message):
        CondaRC(condarc_path)


def test_error_saving_condarc(mocker, tmp_path):
    """
    Testing to make sure the appropriate error is raised when an OSError occurs on file save
    """
    error_message = "Not allowed to write"
    condarc_path = tmp_path / ".condarc"
    condarc_path.touch()
    mock_path = MagicMock()
    mock_path.open.side_effect = [
        condarc_path.open("r"),
        PermissionError(error_message),
    ]

    with pytest.raises(CondaRCError, match=error_message):
        condarc = CondaRC(mock_path)
        condarc.save()



from unittest.mock import MagicMock

import pytest
from conda.exceptions import CondaError
from conda.models.channel import Channel
from keyring.errors import PasswordDeleteError
from requests.auth import _basic_auth_str

from conda_auth.handlers.basic_auth import (
    manager,
    BasicAuthHandler,
    HTTP_BASIC_AUTH_NAME,
    BasicAuthManager,
)
from conda_auth.exceptions import CondaAuthError


@pytest.fixture(autouse=True)
def clean_up_manager_cache():
    """Makes sure the manager cache gets emptied after each test run"""
    manager.cache_clear()


def test_basic_auth_manager_no_previous_secret(keyring):
    """
    Test to make sure when there is no password set, we are able to set a new
    password via the ``getpass`` function.
    """
    settings = {
        "username": "admin",
    }
    channel = Channel("tester")

    # setup mocks
    keyring(None)

    # run code under test
    with pytest.raises(CondaAuthError, match="Password not found"):
        manager.store(channel, settings)


def test_basic_auth_manager_no_secret_or_username(keyring):
    """
    Test to make sure when there is no password or username set, we raise the correct
    exception.
    """
    settings = {}
    channel = Channel("tester")

    # setup mocks
    keyring(None)

    # run code under test
    with pytest.raises(CondaAuthError, match="Username not found"):
        manager.store(channel, settings)


def test_basic_auth_manager_with_previous_secret(keyring):
    """
    Test to make sure when there is a password set, we retrieve it and set the
    cache object appropriately.
    """
    secret = "secret"
    settings = {
        "username": "admin",
    }
    channel = Channel("tester")

    # setup mocks
    keyring(secret)

    # run code under test
    manager.store(channel, settings)

    # make assertions
    assert manager._cache == {channel.canonical_name: ("admin", secret)}


def test_basic_auth_manager_cache_exists(keyring):
    """
    Test to make sure that everything works as expected when a cache entry
    already exists for a credential set.
    """
    secret = "secret"
    username = "admin"
    settings = {
        "username": username,
    }
    channel = Channel("tester")
    manager._cache = {channel.canonical_name: (username, secret)}

    # setup mocks
    keyring_mock, _ = keyring(secret)

    # run code under test
    manager.store(channel, settings)

    # make assertions
    assert manager._cache == {channel.canonical_name: (username, secret)}
    keyring_mock.get_password.assert_not_called()


def test_basic_auth_manager_remove_existing_secret(keyring):
    """
    Test to make sure that removing a password that exist works.
    """
    secret = "secret"
    settings = {
        "username": "username",
    }
    channel = Channel("tester")

    # setup mocks
    keyring_mock, _ = keyring(secret)

    # run code under test
    manager.remove_secret(channel, settings)

    # make assertions
    keyring_mock.delete_password.assert_called_once()


def test_basic_auth_manager_remove_existing_secret_no_username(keyring):
    """
    Test to make sure that when removing a password that exist it fails when no username is present
    """
    secret = "secret"
    settings = {}
    channel = Channel("tester")

    # setup mocks
    keyring(secret)

    # run code under test
    with pytest.raises(CondaAuthError, match="Username not found"):
        manager.remove_secret(channel, settings)


def test_basic_auth_manager_remove_non_existing_secret(keyring):
    """
    Test make sure that when removing a secret that does not exist, the appropriate
    exception and message is raised and shown.
    """
    secret = "secret"
    settings = {
        "username": "username",
    }
    channel = Channel("tester")

    # setup mocks
    keyring_mock, _ = keyring(secret)
    message = "Secret not found."
    keyring_mock.delete_password.side_effect = PasswordDeleteError(message)

    # make assertions
    with pytest.raises(CondaAuthError, match=f"Unable to remove secret: {message}"):
        manager.remove_secret(channel, settings)


def test_basic_auth_handler(keyring):
    """
    Test to make sure that we can successfully instantiate and call the ``BasicAuthHandler``
    """
    channel_name = "channel"
    password = "password"
    username = "username"
    channel = Channel(channel_name)

    # setup mocks
    keyring(None)

    manager.store(channel, {"username": username, "password": password})

    auth_handler = BasicAuthHandler(channel_name)

    request = MagicMock()
    request.headers = {}

    request = auth_handler(request)

    assert request.headers == {"Authorization": _basic_auth_str(username, password)}


def test_basic_auth_handler_equals_methods(keyring):
    """
    Test to make sure that we can instantiate multiple ``BasicAuthHandler`` objects and then
    compare the two objects
    """
    channel_name_one = "channel_two"
    channel_name_two = "channel_two"
    password = "password"
    username = "username"
    channel_one = Channel(channel_name_one)
    channel_two = Channel(channel_name_two)

    # setup mocks
    keyring(None)

    manager.store(channel_one, {"username": username, "password": password})
    manager.store(channel_two, {"username": username, "password": password})

    auth_handler_one = BasicAuthHandler(channel_name_one)
    auth_handler_two = BasicAuthHandler(channel_name_two)

    assert (auth_handler_one == auth_handler_two) is True
    assert (auth_handler_one != auth_handler_two) is False


def test_basic_auth_handler_no_credentials_available_error():
    """
    Test to make sure that we raise an error when no credentials can be found in the application's
    cache
    """
    channel_name = "http://localhost"

    with pytest.raises(
        CondaError,
        match=f"Unable to find user credentials for requests with channel {channel_name}",
    ):
        BasicAuthHandler(channel_name)


def test_token_auth_manager_get_auth_class():
    """
    Simple test to make sure we get the expected type back from the ``get_auth_class``
    method
    """
    assert manager.get_auth_class() is BasicAuthHandler


def test_token_auth_manager_get_auth_type():
    """
    Simple test to make sure we get the expected value back from the ``get_auth_type``
    method
    """
    assert manager.get_auth_type() == HTTP_BASIC_AUTH_NAME


def test_basic_auth_manager_hook_action(keyring):
    """
    Test to make sure we can successfully call the ``hook_action`` method for the
    ``BasicAuthManager``.
    """
    channel = "channel"
    username = "username"
    password = "password"

    # setup mocks
    context = MagicMock()
    context.channels = (channel,)
    context.channel_settings = [
        {"channel": channel, "auth": HTTP_BASIC_AUTH_NAME, "username": username}
    ]
    keyring(password)

    token_manager = BasicAuthManager(context)
    token_manager.hook_action("create")

    assert token_manager._cache == {channel: (username, password)}



from conda_auth.cli import auth, SUCCESSFUL_LOGOUT_MESSAGE
from conda_auth.constants import PLUGIN_NAME
from conda_auth.handlers.basic_auth import HTTP_BASIC_AUTH_NAME
from conda_auth.exceptions import CondaAuthError


def test_logout_of_active_session(mocker, runner, keyring):
    """
    Logs out of currently active session; this essentially just removes the "keyring" entry
    """
    channel_name = "tester"
    secret = "password"
    username = "user"

    # setup mocks
    mock_context = mocker.patch("conda_auth.cli.context")
    keyring_mock, _ = keyring(secret)
    mock_context.channel_settings = [
        {"channel": channel_name, "auth": HTTP_BASIC_AUTH_NAME, "username": username}
    ]

    # run command
    result = runner.invoke(auth, ["logout", channel_name])

    assert SUCCESSFUL_LOGOUT_MESSAGE in result.output
    assert result.exit_code == 0

    # Make sure the delete password call was invoked correctly
    assert keyring_mock.delete_password.mock_calls == [
        mocker.call(f"{PLUGIN_NAME}::{HTTP_BASIC_AUTH_NAME}::{channel_name}", username)
    ]


def test_logout_of_non_existing_session(mocker, runner, keyring):
    """
    Logs out of currently active session; this essentially just removes the "keyring" entry
    """
    channel_name = "tester"

    # setup mocks
    mock_context = mocker.patch("conda_auth.cli.context")
    keyring(None)
    mock_context.channel_settings = []

    # run command
    result = runner.invoke(auth, ["logout", channel_name])
    exc_type, exception, _ = result.exc_info

    assert exc_type == CondaAuthError
    assert "Unable to find information about logged in session." in exception.message



# Copyright (C) 2022 Anaconda, Inc
# Copyright (C) 2023 conda
# SPDX-License-Identifier: BSD-3-Clause
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = html_title = "conda auth"
copyright = "2023, Anaconda, Inc."
author = "Anaconda, Inc."

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.graphviz",
    "sphinx.ext.ifconfig",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.viewcode",
    "sphinxcontrib.mermaid",
    "sphinx_sitemap",
    "sphinx_design",
    "sphinx_copybutton",
]

myst_heading_anchors = 3
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "smartquotes",
    "substitution",
    "tasklist",
]


templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]

# Serving the robots.txt since we want to point to the sitemap.xml file
html_extra_path = ["robots.txt"]

html_theme_options = {
    "github_url": "https://github.com/conda-incubator/conda-auth",
    "collapse_navigation": False,
    "navigation_depth": 1,
    "use_edit_page_button": True,
    "show_toc_level": 1,
    "navbar_align": "left",
    "header_links_before_dropdown": 6,
    "show_nav_level": 2,
    # "announcement": "<p>This is the documentation of the new conda-libmamba-solver plugin!</p>",
}

html_context = {
    "github_user": "conda-incubator",
    "github_repo": "conda-auth",
    "github_version": "main",
    "doc_path": "docs",
}

# We don't have a locale set, so we can safely ignore that for the sitemaps.
sitemap_locales = [None]
# We're hard-coding stable here since that's what we want Google to point to.
sitemap_url_scheme = "{link}"



from __future__ import annotations

from abc import ABC, abstractmethod


class Storage(ABC):
    """ABC class for all credential storage backends"""

    @abstractmethod
    def set_password(self, key_id: str, username: str, password: str) -> None:
        """
        Sets the password for a specific ``key_id``
        """

    @abstractmethod
    def get_password(self, key_id: str, username: str) -> str | None:
        """
        Gets the password for a specific ``key_id``
        """

    @abstractmethod
    def delete_password(self, key_id: str, username: str) -> None:
        """
        Deletes the password for a specific ``key_id``
        """



from __future__ import annotations

from typing import Literal

import click
from conda.base.context import context
from conda.models.channel import Channel

from .condarc import CondaRC, CondaRCError
from .exceptions import CondaAuthError
from .handlers import (
    AuthManager,
    basic_auth_manager,
    token_auth_manager,
    HTTP_BASIC_AUTH_NAME,
    TOKEN_NAME,
)
from .options import ConditionalOption

# Constants
AUTH_MANAGER_MAPPING = {
    HTTP_BASIC_AUTH_NAME: basic_auth_manager,
    TOKEN_NAME: token_auth_manager,
}

SUCCESSFUL_LOGIN_MESSAGE = "Successfully stored credentials"

SUCCESSFUL_LOGOUT_MESSAGE = "Successfully removed credentials"

SUCCESSFUL_COLOR = "green"

FAILURE_COLOR = "red"

OPTION_DEFAULT = "CONDA_AUTH_DEFAULT"


def get_auth_manager(
    auth: str | None = None,
    basic: bool | None = None,
    token: str | Literal[False] | None = None,
    **kwargs,
) -> tuple[str, AuthManager]:
    """
    Based on CLI options provided, return the correct auth manager to use.
    """
    if auth:  # set in .condarc
        pass
    elif basic:  # defined on CLI
        auth = HTTP_BASIC_AUTH_NAME
    elif token:  # defined on CLI
        auth = TOKEN_NAME
    else:
        raise CondaAuthError("Missing authentication type.")

    # check if auth defined maps to a valid auth manager
    if not (auth_manager := AUTH_MANAGER_MAPPING.get(auth)):
        raise CondaAuthError(
            "Invalid authentication type. "
            f"Valid types are: {set(AUTH_MANAGER_MAPPING)}"
        )

    return auth, auth_manager


@click.group("auth", context_settings={"help_option_names": ["-h", "--help"]})
def auth():
    """
    Commands for handling authentication within conda
    """


@auth.command("login")
@click.argument("channel", callback=lambda ctx, param, value: Channel(value))
@click.option(
    "-b",
    "--basic",
    help="Save login credentials as HTTP basic authentication",
    cls=ConditionalOption,
    is_flag=True,
    mutually_exclusive={"token"},
    not_required_if={"token"},
)
@click.option(
    "-u",
    "--username",
    help="Username to use for private channels using HTTP Basic Authentication",
    cls=ConditionalOption,
    prompt_when={"basic"},
    mutually_exclusive={"token"},
)
@click.option(
    "-p",
    "--password",
    help="Password to use for private channels using HTTP Basic Authentication",
    cls=ConditionalOption,
    prompt_when={"basic"},
    hide_input=True,
    mutually_exclusive={"token"},
)
@click.option(
    "-t",
    "--token",
    help="Token to use for private channels using an API token",
    cls=ConditionalOption,
    prompt=True,
    prompt_required=False,
    mutually_exclusive={"basic", "username", "password"},
    not_required_if={"basic"},
)
def login(channel: Channel, **kwargs):
    """
    Log in to a channel by storing the credentials or tokens associated with it
    """
    auth_type, auth_manager = get_auth_manager(**kwargs)
    username: str | None = auth_manager.store(channel, kwargs)

    click.echo(click.style(SUCCESSFUL_LOGIN_MESSAGE, fg=SUCCESSFUL_COLOR))

    try:
        condarc = CondaRC()
        if auth_type == TOKEN_NAME:
            username = None
        condarc.update_channel_settings(channel.canonical_name, auth_type, username)
        condarc.save()
    except CondaRCError as exc:
        raise CondaAuthError(str(exc))


@auth.command("logout")
@click.argument("channel", callback=lambda ctx, param, value: Channel(value))
def logout(channel: Channel):
    """
    Log out of a channel by removing any credentials or tokens associated with it.
    """
    settings = next(
        (
            settings
            for settings in context.channel_settings
            if settings.get("channel") == channel.canonical_name
        ),
        None,
    )
    if not settings:
        raise CondaAuthError("Unable to find information about logged in session.")

    auth_type, auth_manager = get_auth_manager(**settings)
    auth_manager.remove_secret(channel, settings)

    click.echo(click.style(SUCCESSFUL_LOGOUT_MESSAGE, fg=SUCCESSFUL_COLOR))



from unittest.mock import MagicMock

import pytest
from conda.exceptions import CondaError
from conda.models.channel import Channel
from keyring.errors import PasswordDeleteError

from conda_auth.exceptions import CondaAuthError
from conda_auth.handlers.token import (
    is_anaconda_dot_org,
    manager,
    USERNAME,
    TOKEN_NAME,
    TokenAuthHandler,
    TokenAuthManager,
)


@pytest.fixture(autouse=True)
def clean_up_manager_cache():
    """Makes sure the manager cache gets emptied after each test run"""
    manager.cache_clear()


@pytest.mark.parametrize(
    "channel_name,expected", (("conda-forge", True), ("http://localhost", False))
)
def test_is_anaconda_dot_org(channel_name, expected):
    """
    Tests the ``is_anaconda_dot_org`` function
    """
    assert is_anaconda_dot_org(channel_name) == expected


def test_token_auth_manager_no_token(mocker, keyring):
    """
    Test to make sure when there is no token set, an exception is raised
    """
    token = "token"
    settings = {}
    channel = Channel("tester")

    # setup mocks
    input_mock = mocker.patch("conda_auth.handlers.token.input")
    input_mock.return_value = token
    keyring(None)

    # run code under test
    with pytest.raises(CondaAuthError, match="Token not found"):
        manager.store(channel, settings)


def test_token_auth_manager_with_token(keyring):
    """
    Test to make sure when there is a token set, we are able to set a new token via the ``input``
    function.
    """
    token = "token"
    settings = {"token": token}
    channel = Channel("tester")

    # setup mocks
    keyring(None)

    # run code under test
    manager.store(channel, settings)

    # make assertions
    assert manager._cache == {channel.canonical_name: (USERNAME, token)}


def test_basic_auth_manager_remove_existing_secret(keyring):
    """
    Test to make sure that removing a password that exist works.
    """
    secret = "secret"
    settings = {
        "username": USERNAME,
    }
    channel = Channel("tester")

    # setup mocks
    keyring_mock, _ = keyring(secret)

    # run code under test
    manager.remove_secret(channel, settings)

    # make assertions
    keyring_mock.delete_password.assert_called_once()


def test_basic_auth_manager_remove_non_existing_secret(keyring):
    """
    Test make sure that when removing a secret that does not exist, the appropriate
    exception and message is raised and shown.
    """
    secret = "secret"
    settings = {
        "username": USERNAME,
    }
    channel = Channel("tester")

    # setup mocks
    keyring_mock, _ = keyring(secret)
    message = "Secret not found."
    keyring_mock.delete_password.side_effect = PasswordDeleteError(message)

    # make assertions
    with pytest.raises(CondaAuthError, match=f"Unable to remove secret: {message}"):
        manager.remove_secret(channel, settings)


def test_token_auth_handler_with_anaconda_dot_org_token(keyring):
    """
    Test to make sure that we can successfully instantiate and call the ``TokenAuthHandler``
    using the anaconda.org formatted token
    """
    channel_name = "channel"
    token = "token"
    channel = Channel(channel_name)

    # setup mocks
    keyring(None)

    manager.store(channel, {"token": token})

    auth_handler = TokenAuthHandler(channel_name)

    request = MagicMock()
    request.headers = {}

    request = auth_handler(request)

    assert request.headers == {"Authorization": f"token {token}"}


def test_token_auth_handler_with_bearer_token(keyring):
    """
    Test to make sure that we can successfully instantiate and call the ``TokenAuthHandler``
    using a bearer token.
    """
    channel_name = "http://localhost"
    token = "token"
    channel = Channel(channel_name)

    # setup mocks
    keyring(None)

    manager.store(channel, {"token": token})

    auth_handler = TokenAuthHandler(channel_name)

    request = MagicMock()
    request.headers = {}

    request = auth_handler(request)

    assert request.headers == {"Authorization": f"Bearer {token}"}


def test_token_auth_handler_no_token_available_error():
    """
    Test to make sure that we raise an error when no token can be found in the application's
    cache
    """
    channel_name = "http://localhost"

    with pytest.raises(
        CondaError,
        match=f"Unable to find authorization token for requests with channel {channel_name}",
    ):
        TokenAuthHandler(channel_name)


def test_token_auth_manager_get_auth_class():
    """
    Simple test to make sure we get the expected type back from the ``get_auth_class``
    method
    """
    assert manager.get_auth_class() is TokenAuthHandler


def test_token_auth_manager_get_auth_type():
    """
    Simple test to make sure we get the expected value back from the ``get_auth_type``
    method
    """
    assert manager.get_auth_type() == TOKEN_NAME


def test_token_auth_manager_hook_action(keyring):
    """
    Test to make sure we can successfully call the ``hook_action`` method for the
    ``TokenAuthManager``.
    """
    channel = "channel"
    token = "token"

    # setup mocks
    context = MagicMock()
    context.channels = (channel,)
    context.channel_settings = [
        {
            "channel": channel,
            "auth": TOKEN_NAME,
        }
    ]
    keyring(token)

    token_manager = TokenAuthManager(context)
    token_manager.hook_action("create")

    assert token_manager._cache == {channel: (USERNAME, token)}



import pytest

from conda_auth.cli import auth


def test_auth_wrapper():
    """
    Test to make sure the ``auth_wrapper`` function works.

    It is run with no arguments which will print the help message and raise a ``SystemExit``
    exception.
    """
    with pytest.raises(SystemExit):
        auth([])



"""
Token implementation for the conda auth handler plugin hook
"""
from __future__ import annotations

from collections.abc import Mapping

from conda.models.channel import Channel
from conda.plugins.types import ChannelAuthBase

from ..constants import PLUGIN_NAME
from ..exceptions import CondaAuthError
from ..storage import storage
from .base import AuthManager

TOKEN_PARAM_NAME: str = "token"
"""
Name of the configuration parameter where token information is stored
"""

USERNAME: str = "token"
"""
Placeholder value for username; This is written to the secret storage backend
"""

TOKEN_NAME: str = "token"
"""
Name used to refer to this authentication handler in configuration
"""


class TokenAuthManager(AuthManager):
    def get_keyring_id(self, channel: Channel) -> str:
        return f"{PLUGIN_NAME}::{TOKEN_NAME}::{channel.canonical_name}"

    def _fetch_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> tuple[str, str]:
        """
        Gets the secrets by checking the keyring and then falling back to interrupting
        the program and asking the user for secret.
        """
        # First tried the value we passed in
        token = settings.get(TOKEN_PARAM_NAME)

        if token is None:
            # Try password manager if there was nothing there
            keyring_id = self.get_keyring_id(channel)
            token = storage.get_password(keyring_id, USERNAME)

            if token is None:
                raise CondaAuthError("Token not found")

        return USERNAME, token

    def remove_secret(
        self, channel: Channel, settings: Mapping[str, str | None]
    ) -> None:
        keyring_id = self.get_keyring_id(channel)

        storage.delete_password(keyring_id, USERNAME)

    def get_auth_type(self) -> str:
        return TOKEN_NAME

    def get_config_parameters(self) -> tuple[str, ...]:
        return (TOKEN_PARAM_NAME,)

    def get_auth_class(self) -> type:
        return TokenAuthHandler


manager = TokenAuthManager()


def is_anaconda_dot_org(channel_name: str) -> bool:
    """
    Determines whether the ``channel_name`` is a https://anaconda.org channel
    """
    channel = Channel(channel_name)
    domain_name = "anaconda.org"

    return any(domain_name in url for url in channel.base_urls)


class TokenAuthHandler(ChannelAuthBase):
    """
    Implements token auth that inserts a token as a header for all network request
    in conda for the channel specified on object instantiation.

    We make a special exception for anaconda.org and set the Authentication header as:

        Authentication: token <token>

    In all other cases, we use the "bearer" format:

        Authentication: Bearer <token>
    """

    def __init__(self, channel_name: str):
        _, self.token = manager.get_secret(channel_name)
        self.is_anaconda_dot_org = is_anaconda_dot_org(channel_name)

        if self.token is None:
            raise CondaAuthError(
                f"Unable to find authorization token for requests with channel {channel_name}"
            )

        super().__init__(channel_name)

    def __call__(self, request):
        if self.is_anaconda_dot_org:
            request.headers["Authorization"] = f"token {self.token}"
        else:
            request.headers["Authorization"] = f"Bearer {self.token}"

        return request
